#  Copyright (C) 2022
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Alexander Medvedev <a.medvedev@abm-jsc.ru>

import hashlib
import logging
import os
import time
from enum import unique, StrEnum
from typing import Optional

from aiohttp import BasicAuth, hdrs, ClientResponse
from aiohttp.web_exceptions import HTTPUnauthorized
from container_tools import conditional_setitem
from http_tools.calculate_hash import HashAlgorithm, calculate_hash
from http_tools.charset import Charset

logger = logging.getLogger(__name__)


def to_bytes(value: str, charset: Optional[Charset] = Charset.UTF8) -> bytes:
    return value.encode(charset)


@unique
class DigestHeaderKey(StrEnum):
    realm = 'realm'
    nonce = 'nonce'
    qop = 'qop'
    algorithm = 'algorithm'
    opaque = 'opaque'


class DigestAuth(BasicAuth):
    _DIGEST_PREFIX = 'Digest '

    def __init__(self, username: str, password: str) -> None:
        self._username = username
        self._password = password

        self._last_nonce = ''
        self._nonce_count = 0

        self._response: Optional[ClientResponse] = None

        self.is_first_request = True

    def _parse_digest_header(self) -> dict[str, str]:
        header_value = self._response.headers.get(hdrs.WWW_AUTHENTICATE)
        if header_value is None or not header_value.startswith(self._DIGEST_PREFIX):
            logger.error(f'Invalid header: {header_value}')
            raise HTTPUnauthorized()
        header_value = header_value[len(self._DIGEST_PREFIX):]

        result = {}
        for key_to_val in header_value.split(', '):
            key, val = key_to_val.split('=')
            result[key] = val.replace('"', '')
        return result

    def set_response_context(self, response: ClientResponse) -> None:
        self._response = response
        self.is_first_request = False

    def encode(self) -> str:
        """Encode credentials."""
        return self.build_digest_header() if not self.is_first_request else ''

    def build_digest_header(self) -> Optional[str]:
        header_key_to_val = self._parse_digest_header()

        realm = header_key_to_val[DigestHeaderKey.realm]
        nonce = header_key_to_val[DigestHeaderKey.nonce]
        qop = header_key_to_val.get(DigestHeaderKey.qop)
        algorithm = header_key_to_val.get(DigestHeaderKey.algorithm, HashAlgorithm.md5).upper()
        opaque = header_key_to_val.get(DigestHeaderKey.opaque)

        a_1 = f'{self._username}:{realm}:{self._password}'
        a_2 = f'{self._response.method}:{self._response.url.path_qs}'

        hash_algorithm = HashAlgorithm(algorithm)
        ha_1 = calculate_hash(to_bytes(a_1), hash_algorithm)
        ha_2 = calculate_hash(to_bytes(a_2), hash_algorithm)

        self._nonce_count = self._nonce_count + 1 if nonce == self._last_nonce else 1
        self._last_nonce = nonce

        cnonce = b''.join([to_bytes(str(self._nonce_count)), to_bytes(nonce), to_bytes(time.ctime()), os.urandom(8)])
        hashed_cnonce = hashlib.sha1(cnonce).hexdigest()[:16]

        if hash_algorithm == HashAlgorithm.md5_sess:
            ha_1 = calculate_hash(to_bytes(f'{ha_1}:{nonce}:{hashed_cnonce}'), hash_algorithm)

        nc_value = f'{self._nonce_count:08x}'
        digest_auth = 'auth'
        if not qop:
            hashed_response = calculate_hash(to_bytes(f'{ha_1}:{{{nonce}:{ha_2}}}'), hash_algorithm)
        elif qop == digest_auth or digest_auth in qop.split(','):
            nonce_bit = f'{nonce}:{nc_value}:{hashed_cnonce}:{digest_auth}:{ha_2}'
            hashed_response = calculate_hash(to_bytes(f'{ha_1}:{nonce_bit}'), hash_algorithm)
        else:
            logger.error(f'Digest header building failed. Incoming qop: {qop} is not correct')
            raise HTTPUnauthorized()

        conditional_setitem(key_to_value := {},
                            username=self._username,
                            realm=realm,
                            nonce=nonce,
                            uri=self._response.url.path_qs,
                            response=hashed_response,
                            opaque=opaque,
                            algorithm=hash_algorithm)
        digest_header_value = ', '.join([f'{key}="{value}"' for key, value in key_to_value.items()])
        if qop:
            digest_header_value += f', qop="{digest_auth}", nc={nc_value}, cnonce="{hashed_cnonce}"'

        return f'{self._DIGEST_PREFIX}{digest_header_value}'



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import abc
from typing import Optional, Mapping, Any

from custom_json import Jsonable

from ..abm_service_connector import AbmServiceConnector


class AbmServiceAuthedConnector(abc.ABC):
    Config = AbmServiceConnector.Config
    Context = AbmServiceConnector.Context

    def __init__(self, config: Config, context: Context) -> None:
        self._connector = AbmServiceConnector(config, context)

    @abc.abstractmethod
    async def get(self,
                  path: str,
                  url_query: Optional[Mapping[str, Any]] = None,
                  headers: Optional[Mapping[str, str]] = None,
                  token: str = None,
                  timeout_sec: Optional[float] = None) -> Any:
        pass

    @abc.abstractmethod
    async def head(self,
                   path: str,
                   url_query: Optional[Mapping[str, Any]] = None,
                   headers: Optional[Mapping[str, str]] = None,
                   token: str = None,
                   timeout_sec: Optional[float] = None) -> Any:
        pass

    @abc.abstractmethod
    async def put(self,
                  path: str,
                  payload: bytes | Jsonable,
                  url_query: Optional[Mapping[str, Any]] = None,
                  headers: Optional[Mapping[str, str]] = None,
                  token: str = None,
                  timeout_sec: Optional[float] = None) -> Any:
        pass

    @abc.abstractmethod
    async def post(self,
                   path: str,
                   payload: bytes | Jsonable,
                   url_query: Optional[Mapping[str, Any]] = None,
                   headers: Optional[Mapping[str, str]] = None,
                   token: str = None,
                   timeout_sec: Optional[float] = None) -> Any:
        pass

    @abc.abstractmethod
    async def patch(self,
                    path: str,
                    payload: bytes | Jsonable,
                    url_query: Optional[Mapping[str, Any]] = None,
                    headers: Optional[Mapping[str, str]] = None,
                    token: str = None,
                    timeout_sec: Optional[float] = None) -> Any:
        pass

    @abc.abstractmethod
    async def delete(self,
                     path: str,
                     payload: bytes | Jsonable,
                     url_query: Optional[Mapping[str, Any]] = None,
                     headers: Optional[Mapping[str, str]] = None,
                     token: str = None,
                     timeout_sec: Optional[float] = None) -> Any:
        pass



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import asyncio
import dataclasses
import logging
from typing import Awaitable, Callable, Iterable, Optional

from aiohttp import web
from aiohttp.web_urldispatcher import UrlDispatcher

from . import HttpStatusCode
from .answer import JsonableAnswer
from .charset import Charset

logger = logging.getLogger(__name__)

Handler = Callable[[web.Request], Awaitable[web.Response]]

REDIRECT_PERIOD = 1
buffer = {"id": 1, 'tasks': {}}


def get_task_id() -> int:
    id_ = buffer['id']
    buffer['id'] += 1
    return id_


def first(iterable: Iterable):
    return next(iter(iterable))


class AskLaterAnswer(JsonableAnswer):
    @dataclasses.dataclass
    class AskLaterPayload:
        task_id: int
        done: bool = False

    def __init__(self, task_id: int, headers: Optional[dict[str, str]] = None, charset: Charset = Charset.UTF8) -> None:
        super().__init__(payload=self.AskLaterPayload(task_id=task_id), headers=headers, charset=charset)

    @classmethod
    def get_class_payload_type(cls) -> type:
        return cls.AskLaterPayload

    @classmethod
    def get_class_status_code(cls) -> HttpStatusCode:
        return HttpStatusCode.OK


def return_result_by_task_id(task_id: int):
    task: asyncio.Task = buffer['tasks'][task_id]
    if task.done():
        print('task is done')
        del buffer['tasks'][task_id]
        return task.result()
    else:
        print('task is not done')
        return AskLaterAnswer(task_id=task_id)


@web.middleware
async def error_middleware(request: web.Request, handler: Handler, _: UrlDispatcher) -> web.Response:
    task_id = request.headers.get('X-TASK-ID')
    if task_id:
        print('delayed task detected')
        return return_result_by_task_id(int(task_id))

    done, pending = await asyncio.wait([handler(request)], timeout=REDIRECT_PERIOD)
    if done:
        print(f"done in first cycle")
        task: asyncio.Task = first(done)
        return task.result()
    else:
        task_id = get_task_id()
        buffer['tasks'][task_id] = first(pending)
        print(f"saved task to buffer: {buffer['tasks']}")
        return AskLaterAnswer(task_id=task_id)
